#!/usr/bin/env python

import bootstrap
import SocketServer
import SimpleHTTPServer
import threading
import os
import os.path
import time
import sys
import getopt
import urllib
import handlers

"""JCapture Framework

This module provides some building blocks for the Jahshaka Capture framework.

The basic structure of the capture functionality is simply that each capture
device is wrapped by an http server, and clients connect to this server to 
receive the video streams and interact with the device - hence, concurrent 
playback and capture can be carried out via two simultaneous connections 
or the client can combine both via a single connection.

Device specific controls (such as transport and tuner controls) are introduced
via the extension mechanism - this essentially provides a means to have 
specific http GET requests delegated to an external python module.

Starting the Server:

Before we can connect to it, we need to start it. 

For a full list of current options, run:

$ jcapture --help

The simplest execution is in conjunction with Hauppage PVR cards (using the 
ivtv driver):

$ jcapture

or:

$ jcapture --port=51030 --device=/dev/video0 --uri=/jah.mpg --mime=video/mpeg

This starts a forked server on port 51030.

To test it, run:

$ jbatch http://localhost:5130/jah.mpg

If all is well, you should be playing out whatever the card is receiving using 
the established settings on your card.

For a dv1394 PAL device, run:

$ jcapture --port=51031 --device=/dev/dv1394/0 --uri=/jah.dv --mime=video/dv --block=144000 --count=25

For a dv1394 NTSC device, run:

$ jcapture --port=51031 --device=/dev/dv1394/0 --uri=/jah.dv --mime=video/dv --block=120000 --count=30

To test either, run:

$ jbatch http://localhost:5131/jah.dv

TODO:

* Querying server capabilities
"""

class jcap_cache:

	"""Cache component

	This component is used to provide a read/write cache for video streams.
	"""

	def __init__( self, block_size = 1024, blocks = 1024 ):
		self.block_size = block_size
		self.blocks = blocks
		self.read_position = -1
		self.write_position = 0
		self.current = [ ]
		self.notify = threading.Lock( )
		self.notification = threading.Condition( self.notify )

	def count( self ):
		return len( self.current )

	def gap( self ):
		result = 0
		if self.read_position != -1:
			if self.read_position <= self.write_position:
				result = self.write_position - self.read_position
			else:
				result = self.write_position - ( self.read_position - self.blocks )
		return result

	def write( self, data ):
		self.notification.acquire( )
		if self.write_position < len( self.current ):
			self.current[ self.write_position ] = data
		else:
			self.current.append( data )
		self.write_position += 1
		if self.write_position >= self.blocks:
			self.write_position = 0
		self.notification.notifyAll( )
		self.notification.release( )
		time.sleep( 0.0001 )

	def read( self ):
		self.notification.acquire( )
		if self.read_position == -1:
			self.read_position = self.write_position
		while self.read_position == self.write_position:
			self.notification.wait( )
		data = self.current[ self.read_position ]
		self.read_position += 1
		if self.read_position >= self.blocks:
			self.read_position = 0
		self.notification.release( )
		return data

	def clear( self ):
		self.read_position = -1
		self.write_position = 0
		self.current = [ ]

class jcap_capture( threading.Thread ):

	"""Basic capture component.

	This class is suitable for devices and fifos which provide blocking file
	access to a video stream. It loops reading from the device and writing 
	into the cache.

	Note that the cache block_size must be set to appropriately for standard DV
	type devices to ensure that clients receive complete frames (144000 for PAL,
	120000 for NTSC and 288000 for DVCPro). MPEG video and audio streams block 
	sizes are arbitrary.
	"""

	def __init__( self, cache, device ):
		threading.Thread.__init__( self )
		self.setDaemon( True )
		self.cache = cache
		self.device = device
		self.running = False

	def run( self ):
		self.running = True
		jcap = None
		if self.device.startswith( "oml:" ):
			jcap = handlers.url_callback( self.device, 0 )
			jcap.open( self.device, 0 )
		else:
			jcap = open( self.device, "rb" )
		try:
			chunk = jcap.read( self.cache.block_size )
			while self.running:
				self.cache.write( chunk )
				chunk = jcap.read( self.cache.block_size )
				while self.running and self.cache.read_position >= 0 and self.cache.read_position != self.cache.write_position:
					time.sleep( 0.00001 )
		finally:
			self.running = False
			jcap.close( )

class jcap_despatch( threading.Thread ):

	"""Despatching Mechanism

	All clients are redirected to here when they request the specified uri.

	Essentially, the despatch mechanism holds a collection of file descriptors
	which are fed the stream coming out of the cache in sequence. Clients can
	disconnect independently, and they'll automatically drop out of the list.
	"""
	
	def __init__( self, server, uri = "/jah.mpg", mime = "video/mpeg" ):
		threading.Thread.__init__( self )
		self.setDaemon( True )
		self.uri = uri
		self.mime = mime
		self.server = server
		self.capture = server.capture
		self.cache = server.cache
		self.clients = [ ]
		self.notify = threading.Lock( )
		self.notification = threading.Condition( self.notify )
		self.running = True
		self.wait = True

	def run( self ):
		self.running = True

		while self.running:
			self.notify.acquire( )
			while self.running and self.wait and len( self.clients ) == 0:
				self.notification.wait( )
			if self.running and not self.capture.running:
				self.capture.start( )
				self.wait = False
			self.notify.release( )
			if self.running:
				chunk = self.cache.read( )
				self.notify.acquire( )
				i = len( self.clients ) - 1
				while i >= 0:
					lock, cond, file = self.clients[ i ]
					try:
						file.write( chunk )
						#file.flush( )
					except Exception, e:
						self.clients.pop( i )
						lock.acquire( )
						cond.notify( )
						lock.release( )
					i -= 1
				self.notify.release( )
				if self.running and self.cache.read_position == self.cache.write_position:
					time.sleep( 0.0001 )

		self.capture.running = False
		for lock, cond, file in self.clients:
			lock.acquire( )
			cond.notify( )
			lock.release( )
		self.clients = [ ]

	def serve( self, wfile ):
		wfile.write( "HTTP/1.1 200 OK\r\n" )
		wfile.write( "Content-type: %s\r\n" % self.mime )
		wfile.write( "\r\n" )
		lock = threading.Lock( )
		cond = threading.Condition( lock )
		lock.acquire( )
		self.notify.acquire( )
		self.clients.append( ( lock, cond, wfile ) )
		self.notification.notifyAll( )
		self.notify.release( )
		cond.wait( )
		lock.release( )

class jcap_handler( SimpleHTTPServer.SimpleHTTPRequestHandler ):

	"""The HTTP Request Handler

	This class is used for all connections - the GET method is overridden
	to allow redirection to the despatch mechanism and any other specific
	methods associated to the device (transport controls, status thread 
	info etc).
	"""

	def do_GET( self ):
		if self.path == "/" and jcap_server.index != "":
			self.path = jcap_server.index
		self.split = self.path.split( "?" )
		self.args = {}
		for i in self.split[ 1: ]:
			pair = i.split( "=" )
			if len( pair ) == 1:
				self.args[ pair[ 0 ] ] = None
			else:
				self.args[ pair[ 0 ] ] = urllib.unquote_plus( pair[ 1 ] )
		if self.split[ 0 ] in self.server.extension_uri:
			self.server.extension.process( self )
		elif self.path == self.server.despatch.uri:
			self.server.despatch.serve( self.wfile )
		elif self.path == "/jah/info":
			self.wfile.write( "HTTP/1.1 200 OK\r\n\r\n" )
			self.wfile.write( "capture=" + self.server.uri + "\r\n" )
			self.wfile.write( "mime=" + self.server.mime + "\r\n" )
			self.wfile.write( "browse=" + str( [ 0, 1 ][ self.server.index != "" ] ) + "\r\n" )
			self.wfile.write( "title=" + self.server.title + "\r\n" )
			if self.server.extension is not None:
				for k in self.server.extension.uris.keys( ):
					self.wfile.write( k + "=" + self.server.extension.uris[ k ] + "\r\n" )
		elif self.path == "/jah/reset":
			self.server.reset( )
			self.wfile.write( "HTTP/1.1 204 OK\r\n\r\n" )
		elif self.path == "/jah/quit":
			self.server.running = False
			self.wfile.write( "HTTP/1.1 204 OK\r\n\r\n" )
		else:
			self.protocol_version = "HTTP/1.1"
			SimpleHTTPServer.SimpleHTTPRequestHandler.do_GET( self )

class jcap_server( SocketServer.ThreadingTCPServer ):

	""" The HTTP Server

	This class handles the instance of the server required.
	"""

	host = ""
	port = 51030
	device = "/dev/video0"
	uri = "/jah.mpg"
	mime = "video/mpeg"
	index = ""
	block_size = 10240
	blocks = 1024
	capture = jcap_capture
	handler = jcap_handler
	despatch = jcap_despatch
	extension = ""
	root = None
	title = "JCapture Server"

	def __init__( self ):
		self.allow_reuse_address = True
		SocketServer.ThreadingTCPServer.__init__( self, ( jcap_server.host, jcap_server.port ), jcap_server.handler )
		self.extension_uri = [ ]
		self.extension = None
		self.load_extension( )
		self.running = True
		self.daemon_threads = True
		self.cache = jcap_cache( jcap_server.block_size, jcap_server.blocks )
		self.startup( )

	def load_extension( self ):
		if jcap_server.extension != "":
			exec "import " + jcap_server.extension + " as extension"
			self.extension = extension.jcap_extension( )
			self.extension_uri = extension.jcap_extension( ).uris.values( )

	def startup( self ):
		self.capture = jcap_capture( self.cache, jcap_server.device )
		self.despatch = jcap_despatch( self, jcap_server.uri, jcap_server.mime )
		self.despatch.start( )

	def serve_forever( self ):
		while self.running:
			self.handle_request( )
			time.sleep( 0.0001 )

	def reset( self ):
		self.despatch.running = False
		self.cache.clear( )
		time.sleep( 5 )
		self.startup( )

def usage( ):
	print """
jcapture [ options ]

Options:

    -b, --block=size: block size in bytes (default: 1024)
    -c, --count=count: number of blocks (default: 1024)
    -d, --device=device: device (default: "/dev/video0")
    -e, --extension=module: jcapture extension module (default: "")
    -h, --host=host: host name (default: "")
    -i, --index=uri: default uri for browser access (default: "/")
    -m, --mime=mime: mime type to use (default: "video/mpeg")
    -p, --port=port: port number (default: 51030)
    -r, --root=dir: root directory for http server (default: current dir)
    -t, --title=title: name of server (default: "JCapture Server")
    -u, --uri=uri: uri to deliver stream on (default: "/jah.mpg")

Special:

    --debug: don't fork
    --help: this information
"""

def main( ):
	try:
		opts, args = getopt.getopt( sys.argv[ 1: ], "bcdehimprtu:", [ "block=", "count=", "device=", "extension=", "host=", "index=", "mime=", "port=", "root=", "title=", "uri=", "debug", "help" ] )
	except getopt.GetoptError:
		print "invalid args"
		usage( )
		sys.exit( 1 )

	fork = hasattr( os, "fork" )

	for o, a in opts:
		if o in ( "-h", "--host" ):
			jcap_server.host = a
		if o in ( "-p", "--port" ):
			jcap_server.port = int( a )
		if o in ( "-d", "--device" ):
			jcap_server.device = a
		if o in ( "-u", "--uri" ):
			jcap_server.uri = a
		if o in ( "-m", "--mime" ):
			jcap_server.mime = a
		if o in ( "-b", "--block" ):
			jcap_server.block_size = int( a )
		if o in ( "-c", "--count" ):
			jcap_server.blocks = int( a )
		if o in ( "-e", "--extension" ):
			jcap_server.extension = a
		if o in ( "-i", "--index" ):
			jcap_server.index = a
		if o in ( "-r", "--root" ):
			jcap_server.root = a
		if o in ( "-t", "--title" ):
			jcap_server.title = a
		if o in ( "--debug" ):
			fork = False
		if o in ( "--help" ):
			usage( )
			sys.exit( )

	if fork:
		pid = os.fork( )
		if pid != 0:
			sys.exit( 0 )

	httpd = jcap_server( )
	httpd.serve_forever( )

if __name__ == "__main__":
	main( )

